# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
TensorFlow-style RNN cell wrappers.
"""

from typing import Callable, Generic, List, Optional, Tuple, TypeVar, Union

import torch
import torch.nn.functional as F
from torch import nn

from texar.torch.core.attention_mechanism import (
    AttentionMechanism, AttentionWrapperState, compute_attention)
from texar.torch.utils import utils
from texar.torch.utils.types import MaybeList

__all__ = [
    'RNNState',
    'LSTMState',
    'HiddenState',
    'wrap_builtin_cell',
    'RNNCellBase',
    'RNNCell',
    'GRUCell',
    'LSTMCell',
    'DropoutWrapper',
    'ResidualWrapper',
    'HighwayWrapper',
    'MultiRNNCell',
    'AttentionWrapper',
]

State = TypeVar('State')
RNNState = torch.Tensor
LSTMState = Tuple[torch.Tensor, torch.Tensor]

HiddenState = MaybeList[Union[RNNState, LSTMState]]


def wrap_builtin_cell(cell: nn.RNNCellBase):
    r"""Convert a built-in :torch_nn:`RNNCellBase` derived RNN cell to
    our wrapped version.

    Args:
        cell: the RNN cell to wrap around.

    Returns:
        The wrapped cell derived from
        :class:`texar.torch.core.cell_wrappers.RNNCellBase`.
    """
    # convert cls to corresponding derived wrapper class
    if isinstance(cell, nn.RNNCell):
        self = RNNCellBase.__new__(RNNCell)
    elif isinstance(cell, nn.GRUCell):
        self = RNNCellBase.__new__(GRUCell)
    elif isinstance(cell, nn.LSTMCell):
        self = RNNCellBase.__new__(LSTMCell)
    else:
        raise TypeError(f"Unrecognized class {type(cell)}.")
    RNNCellBase.__init__(self, cell)
    return self


class RNNCellBase(nn.Module, Generic[State]):
    r"""The base class for RNN cells in our framework. Major differences over
    :torch_nn:`RNNCell` are two-fold:

    1. Holds an :torch_nn:`Module` which could either be a built-in
       RNN cell or a wrapped cell instance. This design allows
       :class:`RNNCellBase` to serve as the base class for both vanilla
       cells and wrapped cells.

    2. Adds :meth:`zero_state` method for initialization of hidden states,
       which can also be used to implement batch-specific initialization
       routines.
    """

    def __init__(self, cell: Union[nn.RNNCellBase, 'RNNCellBase']):
        super().__init__()
        if not isinstance(cell, nn.Module):
            raise ValueError("Type of parameter 'cell' must be derived from"
                             "nn.Module, and has 'input_size' and 'hidden_size'"
                             "attributes.")
        self._cell = cell

    @property
    def input_size(self) -> int:
        r"""The number of expected features in the input."""
        return self._cell.input_size

    @property
    def hidden_size(self) -> int:
        r"""The number of features in the hidden state."""
        return self._cell.hidden_size

    @property
    def _param(self) -> nn.Parameter:
        r"""Convenience method to access a parameter under the module. Useful
        when creating tensors of the same attributes using `param.new_*`.
        """
        return next(self.parameters())

    def init_batch(self):
        r"""Perform batch-specific initialization routines. For most cells this
        is a no-op.
        """
        pass

    def zero_state(self, batch_size: int) -> State:
        r"""Return zero-filled state tensor(s).

        Args:
            batch_size: int, the batch size.

        Returns:
            State tensor(s) initialized to zeros. Note that different subclasses
            might return tensors of different shapes and structures.
        """
        self.init_batch()
        if isinstance(self._cell, nn.RNNCellBase):
            state = self._param.new_zeros(
                batch_size, self.hidden_size, requires_grad=False)
        else:
            state = self._cell.zero_state(batch_size)
        return state

    def forward(self,  # type: ignore
                input: torch.Tensor, state: Optional[State] = None) \
            -> Tuple[torch.Tensor, State]:
        r"""
        Returns:
            A tuple of (output, state). For single layer RNNs, output is
            the same as state.
        """
        if state is None:
            batch_size = input.size(0)
            state = self.zero_state(batch_size)
        return self._cell(input, state)


class BuiltinCellWrapper(RNNCellBase[State]):
    r"""Base class for wrappers over built-in :torch_nn:`RNNCellBase`
    RNN cells.
    """

    def forward(self,  # type: ignore
                input: torch.Tensor, state: Optional[State] = None) \
            -> Tuple[torch.Tensor, State]:
        if state is None:
            batch_size = input.size(0)
            state = self.zero_state(batch_size)
        new_state = self._cell(input, state)
        return new_state, new_state


class RNNCell(BuiltinCellWrapper[RNNState]):
    r"""A wrapper over :torch_nn:`RNNCell`."""

    def __init__(self, input_size, hidden_size, bias=True, nonlinearity="tanh"):
        cell = nn.RNNCell(
            input_size, hidden_size, bias=bias, nonlinearity=nonlinearity)
        super().__init__(cell)


class GRUCell(BuiltinCellWrapper[RNNState]):
    r"""A wrapper over :torch_nn:`GRUCell`."""

    def __init__(self, input_size, hidden_size, bias=True):
        cell = nn.GRUCell(input_size, hidden_size, bias=bias)
        super().__init__(cell)


class LSTMCell(BuiltinCellWrapper[LSTMState]):
    r"""A wrapper over :torch_nn:`LSTMCell`, additionally providing the
    option to initialize the forget-gate bias to a constant value.
    """

    def __init__(self, input_size, hidden_size, bias=True,
                 forget_bias: Optional[float] = None):
        if forget_bias is not None and not bias:
            raise ValueError("Parameter 'forget_bias' must be set to None when"
                             "'bias' is set to False.")
        cell = nn.LSTMCell(input_size, hidden_size, bias=bias)
        if forget_bias is not None:
            with torch.no_grad():
                cell.bias_ih[hidden_size:(2 * hidden_size)].fill_(forget_bias)
                cell.bias_hh[hidden_size:(2 * hidden_size)].fill_(forget_bias)
        super().__init__(cell)

    def zero_state(self, batch_size: int) -> LSTMState:
        r"""Returns the zero state for LSTMs as (h, c)."""
        state = self._param.new_zeros(
            batch_size, self.hidden_size, requires_grad=False)
        return state, state

    def forward(self,  # type: ignore
                input: torch.Tensor, state: Optional[LSTMState] = None) \
            -> Tuple[torch.Tensor, LSTMState]:
        if state is None:
            batch_size = input.size(0)
            state = self.zero_state(batch_size)
        new_state = self._cell(input, state)
        return new_state[0], new_state


class DropoutWrapper(RNNCellBase[State]):
    r"""Operator adding dropout to inputs and outputs of the given cell."""

    def __init__(self, cell: RNNCellBase[State],
                 input_keep_prob: float = 1.0,
                 output_keep_prob: float = 1.0,
                 state_keep_prob: float = 1.0,
                 variational_recurrent=False):
        r"""Create a cell with added input, state, and/or output dropout.

        If `variational_recurrent` is set to `True` (**NOT** the default
        behavior), then the same dropout mask is applied at every step, as
        described in:

        Y. Gal, Z Ghahramani.  "A Theoretically Grounded Application of Dropout
        in Recurrent Neural Networks".  https://arxiv.org/abs/1512.05287

        Otherwise a different dropout mask is applied at every time step.

        Note, by default (unless a custom `dropout_state_filter` is provided),
        the memory state (`c` component of any `LSTMStateTuple`) passing through
        a `DropoutWrapper` is never modified.  This behavior is described in the
        above article.

        Args:
            cell: an RNNCell.
            input_keep_prob: float between 0 and 1, input keep probability;
                if it is constant and 1, no input dropout will be added.
            output_keep_prob: float between 0 and 1, output keep probability;
                if it is constant and 1, no output dropout will be added.
            state_keep_prob: float between 0 and 1, output keep probability;
                if it is constant and 1, no output dropout will be added.
                State dropout is performed on the outgoing states of the cell.
            variational_recurrent: bool.  If `True`, then the same dropout
                pattern is applied across all time steps for one batch. This is
                implemented by initializing dropout masks in :meth:`zero_state`.
        """
        super().__init__(cell)

        for prob, attr in [(input_keep_prob, "input_keep_prob"),
                           (state_keep_prob, "state_keep_prob"),
                           (output_keep_prob, "output_keep_prob")]:
            if prob < 0.0 or prob > 1.0:
                raise ValueError(
                    f"Parameter '{attr}' must be between 0 and 1: {prob:d}")

        self._input_keep_prob = input_keep_prob
        self._output_keep_prob = output_keep_prob
        self._state_keep_prob = state_keep_prob

        self._variational_recurrent = variational_recurrent
        self._recurrent_input_mask: Optional[torch.Tensor] = None
        self._recurrent_output_mask: Optional[torch.Tensor] = None
        self._recurrent_state_mask: Optional[torch.Tensor] = None

    def _new_mask(self, batch_size: int, mask_size: int,
                  prob: float) -> torch.Tensor:
        return self._param.new_zeros(batch_size, mask_size).bernoulli_(prob)

    def init_batch(self):
        r"""Initialize dropout masks for variational dropout.

        Note that we do not create dropout mask here, because the batch size
        may not be known until actual input is passed in.
        """
        self._recurrent_input_mask = None
        self._recurrent_output_mask = None
        self._recurrent_state_mask = None

    def _dropout(self, tensor: torch.Tensor, keep_prob: float,
                 mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        r"""Decides whether to perform standard dropout or recurrent dropout."""
        if keep_prob == 1.0 or not self.training:
            return tensor
        if mask is not None:
            return tensor.mul(mask).mul(1.0 / keep_prob)
        return F.dropout(tensor, 1.0 - keep_prob, self.training)

    def forward(self,  # type: ignore
                input: torch.Tensor, state: Optional[State] = None) \
            -> Tuple[torch.Tensor, State]:
        if self.training and self._variational_recurrent:
            # Create or check recurrent masks.
            batch_size = input.size(0)
            for name, size in [('input', self.input_size),
                               ('output', self.hidden_size),
                               ('state', self.hidden_size)]:
                prob = getattr(self, f'_{name}_keep_prob')
                if prob == 1.0:
                    continue
                mask = getattr(self, f'_recurrent_{name}_mask')
                if mask is None:
                    # Initialize the mask according to current batch size.
                    mask = self._new_mask(batch_size, size, prob)
                    setattr(self, f'_recurrent_{name}_mask', mask)
                else:
                    # Check that size matches.
                    if mask.size(0) != batch_size:
                        raise ValueError(
                            "Variational recurrent dropout mask does not "
                            "support variable batch sizes across time steps")

        input = self._dropout(input, self._input_keep_prob,
                              self._recurrent_input_mask)
        output, new_state = super().forward(input, state)
        output = self._dropout(output, self._output_keep_prob,
                               self._recurrent_output_mask)
        new_state = utils.map_structure(
            lambda x: self._dropout(
                x, self._state_keep_prob, self._recurrent_state_mask),
            new_state)
        return output, new_state


class ResidualWrapper(RNNCellBase[State]):
    r"""RNNCell wrapper that ensures cell inputs are added to the outputs."""

    def forward(self,  # type: ignore
                input: torch.Tensor, state: Optional[State] = None) \
            -> Tuple[torch.Tensor, State]:
        output, new_state = super().forward(input, state)
        output = input + output
        return output, new_state


class HighwayWrapper(RNNCellBase[State]):
    r"""RNNCell wrapper that adds highway connection on cell input and output.

    Based on: `R. K. Srivastava, K. Greff, and J. Schmidhuber, "Highway
    networks", arXiv preprint arXiv:1505.00387, 2015.`
    https://arxiv.org/pdf/1505.00387.pdf
    """

    def __init__(self, cell: RNNCellBase[State],
                 carry_bias_init: Optional[float] = None,
                 couple_carry_transform_gates: bool = True):
        r"""Constructs a `HighwayWrapper` for `cell`.

        Args:
            cell: An instance of `RNNCell`.
            carry_bias_init: float, carry gates bias initialization.
            couple_carry_transform_gates: boolean, should the Carry and
                Transform gate be coupled.
        """
        super().__init__(cell)

        self.carry = nn.Linear(self.input_size, self.input_size)
        if not couple_carry_transform_gates:
            self.transform = nn.Linear(self.input_size, self.input_size)

        self._coupled = couple_carry_transform_gates
        if carry_bias_init is not None:
            nn.init.constant_(self.carry.bias, carry_bias_init)
            if not couple_carry_transform_gates:
                nn.init.constant_(self.transform.bias, -carry_bias_init)

    def forward(self,  # type: ignore
                input: torch.Tensor, state: Optional[State] = None) \
            -> Tuple[torch.Tensor, State]:
        output, new_state = super().forward(input, state)
        carry = torch.sigmoid(self.carry(input))
        if self._coupled:
            transform = 1 - carry
        else:
            transform = torch.sigmoid(self.transform(input))
        output = input * carry + output * transform
        return output, new_state


class MultiRNNCell(RNNCellBase[List[State]]):
    r"""RNN cell composed sequentially of multiple simple cells.

    .. code-block:: python

        sizes = [128, 128, 64]
        cells = [BasicLSTMCell(input_size, hidden_size)
                 for input_size, hidden_size in zip(sizes[:-1], sizes[1:])]
        stacked_rnn_cell = MultiRNNCell(cells)
    """

    _cell: nn.ModuleList  # type: ignore

    def __init__(self, cells: List[RNNCellBase[State]]):
        r"""Create a RNN cell composed sequentially of a number of RNNCells.

        Args:
          cells: list of RNNCells that will be composed in this order.

        Raises:
          ValueError: if cells is empty (not allowed).
        """
        if len(cells) == 0:
            raise ValueError("Parameter 'cells' should not be empty.")
        cell = nn.ModuleList(cells)
        super().__init__(cell)  # type: ignore

    @property
    def input_size(self):
        return self._cell[0].input_size

    @property
    def hidden_size(self):
        return self._cell[-1].hidden_size

    def init_batch(self):
        for cell in self._cell:
            cell.init_batch()

    def zero_state(self, batch_size: int) -> List[State]:
        states = [cell.zero_state(batch_size)  # type: ignore
                  for cell in self._cell]
        return states

    def forward(self,  # type: ignore
                input: torch.Tensor,
                state: Optional[List[State]] = None) \
            -> Tuple[torch.Tensor, List[State]]:
        r"""Run this multi-layer cell on inputs, starting from state."""
        if state is None:
            batch_size = input.size(0)
            state = self.zero_state(batch_size)
        new_states = []
        output = input
        for cell, hx in zip(self._cell, state):
            output, new_state = cell(output, hx)
            new_states.append(new_state)
        return output, new_states


class AttentionWrapper(RNNCellBase[AttentionWrapperState]):
    r"""Wraps another `RNNCell` with attention."""

    def __init__(self,
                 cell: RNNCellBase,
                 attention_mechanism: MaybeList[AttentionMechanism],
                 attention_layer_size: Optional[MaybeList[int]] = None,
                 alignment_history: bool = False,
                 cell_input_fn: Optional[Callable[[torch.Tensor, torch.Tensor],
                                                  torch.Tensor]] = None,
                 output_attention: bool = True):
        r"""Wraps RNN cell with attention.
        Construct the `AttentionWrapper`.

        Args:
            cell: An instance of RNN cell.
            attention_mechanism: A list of
                :class:`~texar.torch.core.AttentionMechanism` instances or a
                single instance.
            attention_layer_size: A list of Python integers or a single Python
                integer, the depth of the attention (output) layer(s). If None
                (default), use the context as attention at each time step.
                Otherwise, feed the context and cell output into the attention
                layer to generate attention at each time step. If
                attention_mechanism is a list, attention_layer_size must be a
                list of the same length.
            alignment_history (bool): whether to store alignment
                history from all time steps in the final output state.
            cell_input_fn (optional): A `callable`.  The default is:
                `lambda inputs, attention: array_ops.concat([inputs, attention],
                -1)`.
            output_attention (bool): If `True` (default), the output at
                each time step is the attention value.  This is the behavior of
                Luong-style attention mechanisms. If `False`, the output at
                each time step is the output of `cell`.  This is the behavior
                of Bahdanau-style attention mechanisms.  In both cases, the
                `attention` tensor is propagated to the next time step via the
                state and is used there. This flag only controls whether the
                attention mechanism is propagated up to the next cell in an RNN
                stack or to the top RNN output.

        Raises:
            TypeError: :attr:`attention_layer_size` is not None and
                `attention_mechanism` is a list but
                :attr:`attention_layer_size` is not; or vice versa.
            ValueError: if `attention_layer_size` is not None,
                :attr:`attention_mechanism` is a list, and its length does not
                match that of :attr:`attention_layer_size`; if
                :attr:`attention_layer_size` and `attention_layer` are set
                simultaneously.
        """
        super().__init__(cell)

        self._is_multi: bool
        if isinstance(attention_mechanism, (list, tuple)):
            self._is_multi = True
            attention_mechanisms = attention_mechanism
            for mechanism in attention_mechanisms:
                if not isinstance(mechanism, AttentionMechanism):
                    raise TypeError(
                        "attention_mechanism must contain only instances of "
                        "AttentionMechanism, saw type: %s" %
                        type(mechanism).__name__)
        else:
            self._is_multi = False
            if not isinstance(attention_mechanism, AttentionMechanism):
                raise TypeError(
                    "attention_mechanism must be an AttentionMechanism or list "
                    "of multiple AttentionMechanism instances, saw type: %s" %
                    type(attention_mechanism).__name__)
            attention_mechanisms = [attention_mechanism]

        if cell_input_fn is None:
            cell_input_fn = (
                lambda inputs, attention: torch.cat((inputs, attention),
                                                    dim=-1))
        else:
            if not callable(cell_input_fn):
                raise TypeError(
                    "cell_input_fn must be callable, saw type: %s" %
                    type(cell_input_fn).__name__)

        self._attention_layers: Optional[nn.ModuleList]

        if attention_layer_size is not None:
            if isinstance(attention_layer_size, (list, tuple)):
                attention_layer_sizes = tuple(attention_layer_size)
            else:
                attention_layer_sizes = (attention_layer_size,)

            if len(attention_layer_sizes) != len(attention_mechanisms):
                raise ValueError(
                    "If provided, attention_layer_size must contain exactly "
                    "one integer per attention_mechanism, saw: %d vs %d"
                    % (len(attention_layer_sizes), len(attention_mechanisms)))

            self._attention_layers = nn.ModuleList(
                nn.Linear(attention_mechanisms[i].encoder_output_size +
                          cell.hidden_size,
                          attention_layer_sizes[i],
                          False) for i in range(len(attention_layer_sizes)))
            self._attention_layer_size = sum(attention_layer_sizes)
        else:
            self._attention_layers = None
            self._attention_layer_size = sum(
                attention_mechanism.encoder_output_size
                for attention_mechanism in attention_mechanisms)

        self._cell = cell
        self.attention_mechanisms = attention_mechanisms
        self._cell_input_fn = cell_input_fn
        self._output_attention = output_attention
        self._alignment_history = alignment_history
        self._initial_cell_state = None

    def _item_or_tuple(self, seq):
        r"""Returns `seq` as tuple or the singular element.
        Which is returned is determined by how the AttentionMechanism(s) were
        passed to the constructor.

        Args:
            seq: A non-empty sequence of items or generator.

        Returns:
            Either the values in the sequence as a tuple if
            AttentionMechanism(s) were passed to the constructor as a sequence
            or the singular element.
        """
        t = tuple(seq)
        if self._is_multi:
            return t
        else:
            return t[0]

    @property
    def output_size(self) -> int:
        r"""The number of features in the output tensor."""
        if self._output_attention:
            return self._attention_layer_size
        else:
            return self._cell.hidden_size

    def zero_state(self,
                   batch_size: int) -> AttentionWrapperState:
        r"""Return an initial (zero) state tuple for this
        :class:`AttentionWrapper`.

        .. note::
                Please see the initializer documentation for details of how
                to call :meth:`zero_state` if using an
                :class:`~texar.torch.core.AttentionWrapper` with a
                :class:`~texar.torch.modules.BeamSearchDecoder`.

        Args:
            batch_size: `0D` integer: the batch size.

        Returns:
            An :class:`~texar.torch.core.AttentionWrapperState` tuple containing
            zeroed out tensors and Python lists.
        """
        cell_state: torch.Tensor = super().zero_state(batch_size)  # type:ignore

        initial_alignments = [None for _ in self.attention_mechanisms]

        alignment_history: List[List[Optional[torch.Tensor]]]
        alignment_history = [[] for _ in initial_alignments]

        return AttentionWrapperState(
            cell_state=cell_state,
            time=0,
            attention=self._param.new_zeros(batch_size,
                                            self._attention_layer_size,
                                            requires_grad=False),
            alignments=self._item_or_tuple(initial_alignments),
            attention_state=self._item_or_tuple(initial_alignments),
            alignment_history=self._item_or_tuple(alignment_history))

    def forward(self,  # type: ignore
                inputs: torch.Tensor,
                state: Optional[AttentionWrapperState],
                memory: torch.Tensor,
                memory_sequence_length: Optional[torch.LongTensor] = None) -> \
            Tuple[torch.Tensor, AttentionWrapperState]:
        r"""Perform a step of attention-wrapped RNN.

        - Step 1: Mix the :attr:`inputs` and previous step's `attention` output
          via `cell_input_fn`.
        - Step 2: Call the wrapped `cell` with this input and its previous
          state.
        - Step 3: Score the cell's output with `attention_mechanism`.
        - Step 4: Calculate the alignments by passing the score through the
          `normalizer`.
        - Step 5: Calculate the context vector as the inner product between the
          alignments and the attention_mechanism's values (memory).
        - Step 6: Calculate the attention output by concatenating the cell
          output and context through the attention layer (a linear layer with
          `attention_layer_size` outputs).

        Args:
            inputs: (Possibly nested tuple of) Tensor, the input at this time
                step.
            state: An instance of
                :class:`~texar.torch.core.AttentionWrapperState` containing
                tensors from the previous time step.
            memory: The memory to query; usually the output of an RNN encoder.
                This tensor should be shaped `[batch_size, max_time, ...]`.
            memory_sequence_length: (optional) Sequence lengths for the batch
                entries in memory.  If provided, the memory tensor rows are
                masked with zeros for values past the respective sequence
                lengths.

        Returns:
            A tuple `(attention_or_cell_output, next_state)`, where

            - `attention_or_cell_output` depending on `output_attention`.
            - `next_state` is an instance of
              :class:`~texar.torch.core.AttentionWrapperState` containing the
              state calculated at this time step.

        Raises:
            TypeError: If `state` is not an instance of
                :class:`~texar.torch.core.AttentionWrapperState`.
        """
        if state is None:
            state = self.zero_state(batch_size=memory.shape[0])
        elif not isinstance(state, AttentionWrapperState):
            raise TypeError("Expected state to be instance of "
                            "AttentionWrapperState. Received type %s instead."
                            % type(state))

        # Step 1: Calculate the true inputs to the cell based on the
        # previous attention value.
        cell_inputs = self._cell_input_fn(inputs, state.attention)
        cell_state = state.cell_state

        cell_output, next_cell_state = self._cell(cell_inputs, cell_state)

        if self._is_multi:
            previous_attention_state = state.attention_state
            previous_alignment_history = state.alignment_history
        else:
            previous_attention_state = [state.attention_state]  # type: ignore
            previous_alignment_history = \
                [state.alignment_history]  # type: ignore

        all_alignments = []
        all_attentions = []
        all_attention_states = []
        maybe_all_histories = []
        for i, attention_mechanism in enumerate(self.attention_mechanisms):
            if previous_attention_state[i] is not None:
                attention_state = previous_attention_state[i]
            else:
                attention_state = attention_mechanism.initial_state(
                    memory.shape[0], memory.shape[1], self._param.dtype,
                    self._param.device)

            attention, alignments, next_attention_state = compute_attention(
                attention_mechanism=attention_mechanism,
                cell_output=cell_output,
                attention_state=attention_state,
                attention_layer=(self._attention_layers[i]
                                 if self._attention_layers else None),
                memory=memory,
                memory_sequence_length=memory_sequence_length)

            if self._alignment_history:
                alignment_history = previous_alignment_history[i] + [alignments]
            else:
                alignment_history = previous_alignment_history[i]

            all_attention_states.append(next_attention_state)
            all_alignments.append(alignments)
            all_attentions.append(attention)
            maybe_all_histories.append(alignment_history)

        attention = torch.cat(all_attentions, 1)
        next_state = AttentionWrapperState(
            time=state.time + 1,
            cell_state=next_cell_state,
            attention=attention,
            attention_state=self._item_or_tuple(all_attention_states),
            alignments=self._item_or_tuple(all_alignments),
            alignment_history=self._item_or_tuple(maybe_all_histories))

        if self._output_attention:
            return attention, next_state
        else:
            return cell_output, next_state



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utils of pre-trained GPT2 tokenizer.

Code structure adapted from:
    `https://github.com/huggingface/pytorch-transformers/blob/master/pytorch_transformers/tokenization_gpt2.py`
"""

from functools import lru_cache

__all__ = [
    "bytes_to_unicode",
    "get_pairs",
]


@lru_cache()
def bytes_to_unicode():
    r"""Returns list of utf-8 byte and a corresponding list of unicode strings.
    The reversible bpe codes work on unicode strings. This means you need a
    large number of unicode characters in your vocab if you want to avoid UNKs.
    When you're at something like a 10B token dataset you end up needing
    around 5K for decent coverage. This is a significant percentage of your
    normal, say, 32K bpe vocab. To avoid that, we want lookup tables between
    utf-8 bytes and unicode strings.

    Note that this function avoids the mapping to whitespace and control
    characters, which is designed specifically for GPT-2 BPE.
    """
    bs = list(range(ord("!"), ord("~") + 1)) + list(
        range(ord("¡"), ord("¬") + 1)) + list(range(ord("®"), ord("ÿ") + 1))
    cs = bs[:]
    n = 0
    for b in range(2**8):
        if b not in bs:
            bs.append(b)
            cs.append(2**8 + n)
            n += 1
    cs = [chr(n) for n in cs]
    return dict(zip(bs, cs))


def get_pairs(word):
    r"""Return set of symbol pairs in a word. Word is represented as tuple of
    symbols (symbols being variable-length strings).

    Example:
        word = "texar"
        get_pairs(word)
        # {('t', 'e'), ('e', 'x'), ('x', 'a'), ('a', 'r')}
    """
    pairs = set()
    prev_char = word[0]
    for char in word[1:]:
        pairs.add((prev_char, char))
        prev_char = char
    return pairs



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utils of pre-trained BERT tokenizer.

Code structure adapted from:
    `https://github.com/huggingface/pytorch-transformers/blob/master/pytorch_transformers/tokenization_bert.py`
"""

from typing import Dict, List, Optional

import collections
import unicodedata


__all__ = [
    "load_vocab",
    "BasicTokenizer",
    "WordpieceTokenizer",
]


def load_vocab(vocab_file: str) -> Dict[str, int]:
    r"""Loads a vocabulary file into a dictionary."""
    vocab: Dict[str, int] = collections.OrderedDict()
    with open(vocab_file, "r", encoding="utf-8") as reader:
        tokens = reader.readlines()
    for index, token in enumerate(tokens):
        token = token.rstrip('\n')
        vocab[token] = index
    return vocab


class BasicTokenizer:
    r"""Runs basic tokenization (punctuation splitting, lower casing, etc.).

    Args:
        do_lower_case: Whether to lower case the input.
        never_split: A list of tokens not to split.
        tokenize_chinese_chars: Whether to tokenize Chinese characters.
            This should likely be deactivated for Japanese:
            see:
            `https://github.com/huggingface/pytorch-pretrained-BERT/issues/328`
    """

    def __init__(self, do_lower_case: bool = True,
                 never_split: Optional[List[str]] = None,
                 tokenize_chinese_chars: bool = True):
        if never_split is None:
            never_split = []
        self.do_lower_case = do_lower_case
        self.never_split = never_split
        self.tokenize_chinese_chars = tokenize_chinese_chars

    def tokenize(self, text: str,
                 never_split: Optional[List[str]] = None) -> \
            List[str]:
        r"""Basic tokenization of a piece of text.

        Split on white spaces only, for sub-word tokenization, see
        WordPieceTokenizer.

        Args:
            text: An input string.
            never_split: A list of tokens not to split.
        """
        never_split = self.never_split + (never_split
                                          if never_split is not None else [])
        text = self._clean_text(text)

        # This was added on November 1st, 2018 for the multilingual and Chinese
        # models. This is also applied to the English models now, but it doesn't
        # matter since the English models were not trained on any Chinese data
        # and generally don't have any Chinese data in them (there are Chinese
        # characters in the vocabulary because Wikipedia does have some Chinese
        # words in the English Wikipedia.).
        if self.tokenize_chinese_chars:
            text = self._tokenize_chinese_chars(text)
        # see: https://github.com/google-research/bert/blob/master/
        # tokenization.py#L201

        orig_tokens = whitespace_tokenize(text)
        split_tokens = []
        for token in orig_tokens:
            if self.do_lower_case and token not in never_split:
                token = token.lower()
                token = self._run_strip_accents(token)
            split_tokens.extend(self._run_split_on_punc(token))

        output_tokens = whitespace_tokenize(" ".join(split_tokens))
        return output_tokens

    @classmethod
    def _run_strip_accents(cls, text: str) -> str:
        r"""Strips accents from a piece of text.

        Example:
            accented_string = 'Málaga'
            _run_strip_accents(accented_string)  # 'Malaga'
        """
        text = unicodedata.normalize("NFD", text)
        output = []
        for char in text:
            cat = unicodedata.category(char)
            if cat == "Mn":
                continue
            output.append(char)
        return "".join(output)

    @classmethod
    def _run_split_on_punc(cls, text: str,
                           never_split: Optional[List[str]] = None) -> \
            List[str]:
        r"""Splits punctuation on a piece of text.

        Example:
            text = 'Texar-PyTorch is an open-source toolkit based on PyTorch.'
            _run_split_on_punc(text)
            # ['Texar', '-', 'PyTorch is an open', '-',
            # 'source toolkit based on PyTorch', '.']
        """
        if never_split is not None and text in never_split:
            return [text]
        chars = list(text)
        i = 0
        start_new_word = True
        output = []
        while i < len(chars):
            char = chars[i]
            if _is_punctuation(char):
                output.append([char])
                start_new_word = True
            else:
                if start_new_word:
                    output.append([])
                start_new_word = False
                output[-1].append(char)
            i += 1

        return ["".join(x) for x in output]

    def _tokenize_chinese_chars(self, text: str) -> str:
        r"""Adds whitespace around any CJK character.

        Example:
            text = '今天天气不错'
            _tokenize_chinese_chars(text)
            # ' 今  天  天  气  不  错 '
        """
        output = []
        for char in text:
            cp = ord(char)
            if self._is_chinese_char(cp):
                output.append(" ")
                output.append(char)
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)

    @classmethod
    def _is_chinese_char(cls, cp: int) -> bool:
        r"""Checks whether cp is the codepoint of a CJK character."""
        # This defines a "chinese character" as anything in the CJK Unicode
        # block:
        #   https://en.wikipedia.org/wiki/CJK_Unified_Ideographs_(Unicode_block)
        #
        # Note that the CJK Unicode block is NOT all Japanese and Korean
        # characters, despite its name. The modern Korean Hangul alphabet is a
        # different block, as is Japanese Hiragana and Katakana. Those
        # alphabets are used to write space-separated words, so they are not
        # treated specially and handled like the all of the other languages.
        if ((0x4E00 <= cp <= 0x9FFF) or
                (0x3400 <= cp <= 0x4DBF) or
                (0x20000 <= cp <= 0x2A6DF) or
                (0x2A700 <= cp <= 0x2B73F) or
                (0x2B740 <= cp <= 0x2B81F) or
                (0x2B820 <= cp <= 0x2CEAF) or
                (0xF900 <= cp <= 0xFAFF) or
                (0x2F800 <= cp <= 0x2FA1F)):
            return True

        return False

    @classmethod
    def _clean_text(cls, text: str) -> str:
        r"""Performs invalid character removal and whitespace cleanup on text.

        Example:
            text = 'Texar-PyTorch\tis an open-source\ntoolkit based on PyTorch.'
            _clean_text(text)
            # 'Texar-PyTorch is an open-source toolkit based on PyTorch.'
        """
        output = []
        for char in text:
            cp = ord(char)
            if cp == 0 or cp == 0xfffd or _is_control(char):
                continue
            if _is_whitespace(char):
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)


class WordpieceTokenizer:
    r"""Runs WordPiece tokenization."""

    def __init__(self, vocab: Dict[str, int],
                 unk_token: str,
                 max_input_chars_per_word: int = 100):
        self.vocab = vocab
        self.unk_token = unk_token
        self.max_input_chars_per_word = max_input_chars_per_word

    def tokenize(self, text: str) -> List[str]:
        r"""Tokenizes a piece of text into its word pieces.

        This uses a greedy longest-match-first algorithm to perform tokenization
        using the given vocabulary.

        For example:
            input = "unaffable"
            output = ["un", "##aff", "##able"]

        Args:
            text: A single token or whitespace separated tokens. This should
                have already been passed through `BasicTokenizer`.

        Returns:
            A list of wordpiece tokens.
        """
        output_tokens = []
        for token in whitespace_tokenize(text):
            assert token is not None
            chars = list(token)
            if len(chars) > self.max_input_chars_per_word:
                output_tokens.append(self.unk_token)
                continue

            is_bad = False
            start = 0
            sub_tokens = []
            while start < len(chars):
                end = len(chars)
                cur_substr = None
                while start < end:
                    substr = "".join(chars[start:end])
                    if start > 0:
                        substr = "##" + substr
                    if substr in self.vocab:
                        cur_substr = substr
                        break
                    end -= 1
                if cur_substr is None:
                    is_bad = True
                    break
                sub_tokens.append(cur_substr)
                start = end

            if is_bad:
                output_tokens.append(self.unk_token)
            else:
                output_tokens.extend(sub_tokens)
        return output_tokens


def whitespace_tokenize(text: str) -> List[str]:
    r"""Runs basic whitespace cleaning and splitting on a piece of text."""
    text = text.strip()
    if not text:
        return []
    tokens: List[str] = text.split()
    return tokens


def _is_whitespace(char: str) -> bool:
    r"""Checks whether `char` is a whitespace character.

    Note: this function is not standard and should be considered for BERT
    tokenization only. See the comments for more details.
    """
    # \t, \n, and \r are technically control characters but we treat them
    # as whitespace since they are generally considered as such.
    if char in (" ", "\t", "\n", "\r"):
        return True
    cat = unicodedata.category(char)
    if cat == "Zs":
        return True
    return False


def _is_control(char: str) -> bool:
    r"""Checks whether `char` is a control character.

    Note: this function is not standard and should be considered for BERT
    tokenization only. See the comments for more details.
    """
    # These are technically control characters but we count them as whitespace
    # characters.
    if char in ("\t", "\n", "\r"):
        return False
    cat = unicodedata.category(char)
    if cat.startswith("C"):
        return True
    return False


def _is_punctuation(char: str) -> bool:
    r"""Checks whether `char` is a punctuation character.

    Note: this function is not standard and should be considered for BERT
    tokenization only. See the comments for more details.
    """
    cp = ord(char)
    # We treat all non-letter/number ASCII as punctuation.
    # Characters such as "^", "$", and "`" are not in the Unicode
    # Punctuation class but we treat them as punctuation anyways, for
    # consistency.
    if ((33 <= cp <= 47) or (58 <= cp <= 64) or
            (91 <= cp <= 96) or (123 <= cp <= 126)):
        return True
    cat = unicodedata.category(char)
    if cat.startswith("P"):
        return True
    return False



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Actions for the Executor module.
"""

from abc import ABC, abstractmethod

from texar.torch.run.executor import Executor


__all__ = [
    "reset_params",
    "scale_lr",
    "early_stop",
]


class Action(ABC):
    @abstractmethod
    def __call__(self, executor: Executor) -> None:
        raise NotImplementedError


class reset_params(Action):
    def __init__(self, training_state: bool = True):
        self.load_training_state = training_state

    def __call__(self, executor: Executor):
        # TODO: Only optimizer?
        executor.load(load_training_state=self.load_training_state)


class scale_lr(Action):
    def __init__(self, scale: float):
        # TODO: Change to accept LRScheduler.
        self.scale = scale

    def __call__(self, executor: Executor):
        new_lr = []
        assert executor.optimizer is not None
        for group in executor.optimizer.param_groups:  # type: ignore
            lr = group['lr'] * self.scale
            group['lr'] = lr
            new_lr.append(lr)

        lr_str = ", ".join(f"{lr:.4e}" for lr in new_lr)
        if len(new_lr) > 1:
            lr_str = f"[{lr_str}]"
        executor.write_log(f"Learning rate scaled by {self.scale}. "
                           f"New LR: {lr_str}")


class early_stop(Action):
    def __init__(self, patience: int):
        if not isinstance(patience, int) or patience <= 0:
            raise ValueError("`patience` must be a positive integer")
        self.patience = patience
        self.count = 0

    def __call__(self, executor: Executor):
        self.count += 1
        if self.count >= self.patience:
            executor.terminate()
            executor.write_log(f"Early stopping patience exhausted, "
                               f"terminating training...")
        else:
            executor.write_log(f"Early stopping patience decrease to "
                               f"{self.patience - self.count}")



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for conv encoders.
"""
import unittest

import torch
from torch import nn

from texar.torch.modules.classifiers.conv_classifiers import Conv1DClassifier


class Conv1DClassifierTest(unittest.TestCase):
    """Tests :class:`~texar.torch.modules.Conv1DClassifier` class.
    """

    def test_classifier(self):
        """Tests classification.
        """
        # case 1: default hparams
        inputs = torch.randn(128, 32, 300)
        classifier = Conv1DClassifier(in_channels=inputs.shape[1],
                                      in_features=inputs.shape[2])

        self.assertEqual(len(classifier.layers), 5)
        self.assertIsInstance(classifier.layers[-1], nn.Linear)
        logits, pred = classifier(inputs)
        self.assertEqual(logits.shape, torch.Size([128, 2]))
        self.assertEqual(pred.shape, torch.Size([128]))

        # case 2
        inputs = torch.randn(128, 32, 300)
        hparams = {
            "num_classes": 10,
            "logit_layer_kwargs": {"bias": False}
        }
        classifier = Conv1DClassifier(in_channels=inputs.shape[1],
                                      in_features=inputs.shape[2],
                                      hparams=hparams)
        logits, pred = classifier(inputs)
        self.assertEqual(logits.shape, torch.Size([128, 10]))
        self.assertEqual(pred.shape, torch.Size([128]))

    def test_classifier_with_no_dense_layer(self):
        """Tests classifier with no dense layer.
        """
        inputs = torch.randn(128, 32, 300)
        hparams = {
            "num_dense_layers": 0,
            "num_classes": 10,
            "logit_layer_kwargs": {"bias": False}
        }
        classifier = Conv1DClassifier(in_channels=inputs.shape[1],
                                      in_features=inputs.shape[2],
                                      hparams=hparams)
        logits, pred = classifier(inputs)
        self.assertEqual(logits.shape, torch.Size([128, 10]))
        self.assertEqual(pred.shape, torch.Size([128]))


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Base class for modules.
"""
from abc import ABC
from typing import Any, Dict, List, Optional, Union

from torch import nn

from texar.torch.hyperparams import HParams

__all__ = [
    'ModuleBase',
]


class ModuleBase(nn.Module, ABC):
    r"""Base class inherited by modules that are configurable through
    hyperparameters.

    This is a subclass of :torch_nn:`Module`.

    A Texar module inheriting :class:`~texar.torch.ModuleBase` is
    **configurable through hyperparameters**. That is, each module defines
    allowed hyperparameters and default values. Hyperparameters not
    specified by users will take default values.

    Args:
        hparams (dict, optional): Hyperparameters of the module. See
            :meth:`default_hparams` for the structure and default values.
    """

    def __init__(self, hparams: Optional[Union[HParams,
                                               Dict[str, Any]]] = None):
        super().__init__()
        if not hasattr(self, '_hparams'):
            self._hparams = HParams(hparams, self.default_hparams())
        else:
            # Probably already parsed by subclasses. We rely on subclass
            # implementations to get this right.
            # As a sanity check, we require `hparams` to be `None` in this case.
            if hparams is not None:
                raise ValueError(
                    "`self._hparams` is already assigned, but `hparams` "
                    "argument is not None.")

    @staticmethod
    def default_hparams() -> Dict[str, Any]:
        r"""Returns a `dict` of hyperparameters of the module with default
        values. Used to replace the missing values of input `hparams`
        during module construction.

        .. code-block:: python

            {
                "name": "module"
            }
        """
        return {
            'name': 'module'
        }

    @property
    def trainable_variables(self) -> List[nn.Parameter]:
        r"""The list of trainable variables (parameters) of the module.
        Parameters of this module and all its submodules are included.

        .. note::
            The list returned may contain duplicate parameters (e.g. output
            layer shares parameters with embeddings). For most usages, it's not
            necessary to ensure uniqueness.
        """
        return [x for x in self.parameters() if x.requires_grad]

    @property
    def hparams(self) -> HParams:
        r"""An :class:`~texar.torch.HParams` instance. The hyperparameters
        of the module.
        """
        return self._hparams

    @property
    def output_size(self):
        r"""The feature size of :meth:`forward` output tensor(s),
        usually it is equal to the last dimension value of the output
        tensor size.
        """
        raise NotImplementedError



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for bleu_moses.
"""

import unittest

import numpy as np

from texar.torch.evals.bleu_moses import sentence_bleu_moses, corpus_bleu_moses


class BLEUMosesTest(unittest.TestCase):
    r"""Tests bleu moses.
    """

    def _test_sentence_bleu(self, references, hypothesis, lowercase,
                            true_bleu):
        bleu = sentence_bleu_moses(references, hypothesis, lowercase=lowercase)

        self.assertAlmostEqual(bleu, true_bleu, places=2)

    def test_sentence_strings(self):
        r"""Tests hypothesis as strings.
        """
        hypothesis = \
            "this is a test sentence to evaluate the good bleu score . 词"
        references = ["this is a test sentence to evaluate the bleu score ."]
        self._test_sentence_bleu(
            references, hypothesis, lowercase=False, true_bleu=67.03)

    def test_sentence_list(self):
        r"""Tests hypothesis as a list of tokens.
        """
        hypothesis = \
            "this is a test sentence to evaluate the good bleu score . 词"
        hypothesis = hypothesis.split()
        references = ["this is a test sentence to evaluate the bleu score ."]
        references = [references[0].split()]
        self._test_sentence_bleu(
            references, hypothesis, lowercase=False, true_bleu=67.03)

    def test_sentence_multi_references(self):
        r"""Tests multiple references.
        """
        hypothesis = \
            "this is a test sentence to evaluate the good bleu score . 词"
        references = ["this is a test sentence to evaluate the bleu score .",
                      "this is a test sentence to evaluate the good score ."]
        self._test_sentence_bleu(
            references, hypothesis, lowercase=False, true_bleu=76.12)

    def test_sentence_numpy(self):
        r"""Tests with numpy format.
        """
        hypothesis = \
            "this is a test sentence to evaluate the good bleu score . 词"
        hypothesis = np.array(hypothesis.split())
        references = ["this is a test sentence to evaluate the bleu score .",
                      "this is a test sentence to evaluate the good score ."]
        references = np.array([np.array(r.split()) for r in references])
        self._test_sentence_bleu(
            references, hypothesis, lowercase=False, true_bleu=76.12)

    def _test_corpus_bleu(self, list_of_references, hypotheses, lowercase,
                          return_all, true_bleu):
        bleu = corpus_bleu_moses(list_of_references, hypotheses,
                                 lowercase=lowercase, return_all=return_all)
        if not return_all:
            self.assertAlmostEqual(bleu, true_bleu, places=2)
        else:
            for ret, true in zip(bleu, true_bleu):
                self.assertAlmostEqual(ret, true, places=2)

    def test_corpus_strings(self):
        r"""Tests corpus level BLEU.
        """
        hypotheses = [
            "this is a test sentence to evaluate the good bleu score . 词",
            "i believe that that the script is 词 perfectly correct ."
        ]
        list_of_references = [
            ["this is a test sentence to evaluate the bleu score .",
             "this is a test sentence to evaluate the good score ."],
            ["i believe that the script is perfectly correct .".split()]
        ]
        self._test_corpus_bleu(list_of_references, hypotheses,
                               False, False, 63.02)

        self._test_corpus_bleu(list_of_references, hypotheses,
                               False, True, [63.02, 87.5, 77.3, 60.0, 38.9])


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various reward related functions.
"""

from typing import Optional

import torch

from texar.torch.utils.shapes import mask_sequences
from texar.torch.utils.utils import sequence_mask

__all__ = [
    "discount_reward",
    "_discount_reward_tensor_1d",
    "_discount_reward_tensor_2d",
]


def discount_reward(reward: torch.Tensor,
                    sequence_length: Optional[torch.LongTensor] = None,
                    discount: float = 1.,
                    normalize: bool = False) -> torch.Tensor:
    r"""Computes discounted reward.

    Args:
        reward: A Tensor. Can be 1D with shape `[batch_size]`,
            or 2D with shape `[batch_size, max_time]`.
        sequence_length (optional): A Tensor of shape `[batch_size]`.
            Time steps beyond the respective sequence lengths will be masked.
            Required if :attr:`reward` is 1D.
        discount (float): A scalar. The discount factor.
        normalize (bool): Whether to normalize the discounted reward, by
            `(discounted_reward - mean) / std`. Here `mean` and `std` are
            over all time steps and all samples in the batch.

    Returns:
        A 2D Tensor of the discounted reward.
    """
    if not isinstance(reward, torch.Tensor):
        reward = torch.tensor(reward)
    if (sequence_length is not None and
            not isinstance(sequence_length, torch.Tensor)):
        sequence_length = torch.tensor(
            sequence_length, dtype=torch.int64, device=reward.device)

    tensor_rank = reward.dim()
    if tensor_rank == 1:
        disc_reward = _discount_reward_tensor_1d(
            reward, sequence_length, discount)
    elif tensor_rank == 2:
        disc_reward = _discount_reward_tensor_2d(
            reward, sequence_length, discount)
    else:
        raise ValueError("The dimension of reward can only be 1 or 2.")

    if normalize:
        mu = torch.mean(disc_reward)
        var = torch.std(disc_reward)
        disc_reward = (disc_reward - mu) / (torch.sqrt(var) + 1e-8)

    return disc_reward


def _discount_reward_tensor_1d(reward: torch.Tensor,
                               sequence_length: Optional[torch.LongTensor],
                               discount: float = 1.) -> torch.Tensor:
    r"""Computes discounted reward.

    Args:
        reward: 1D Tensor with shape `[batch_size]`.
        sequence_length: A Tensor of shape `[batch_size]`.
        Time steps beyond the respective sequence lengths will be masked.
        discount (float): A scalar. The discount factor.

    Returns:
        A 2D Tensor of the discounted reward.
    """
    if sequence_length is None:
        raise ValueError('sequence_length must not be `None` for 1D reward.')

    if not isinstance(sequence_length, torch.Tensor):
        sequence_length = torch.tensor(
            sequence_length, dtype=torch.int64, device=reward.device)

    batch_size = reward.shape[0]
    max_seq_length = torch.max(sequence_length)
    dtype: torch.dtype = reward.dtype

    if discount == 1.:
        disc_reward = reward.unsqueeze(-1).expand(batch_size, max_seq_length)
    else:
        mask = sequence_mask(sequence_length, dtype=dtype)
        mask = torch.cat((mask[:, 1:], torch.zeros_like(mask[:, -1:])), dim=1)
        # Make each row = [discount, ..., discount, 1, ..., 1]
        dmat = mask * discount + (1 - mask)
        dmat = torch.flip(dmat, (1,))
        dmat = torch.cumprod(dmat, dim=1)
        dmat = torch.flip(dmat, (1,))
        disc_reward = dmat * reward.unsqueeze(-1)

    disc_reward = mask_sequences(disc_reward, sequence_length, dtype=dtype)

    return disc_reward


def _discount_reward_tensor_2d(
        reward: torch.Tensor,
        sequence_length: Optional[torch.LongTensor] = None,
        discount: float = 1.) -> torch.Tensor:
    r"""Computes discounted reward.

    Args:
        reward: 2D Tensor with shape `[batch_size, max_time]`.
        sequence_length (optional): A Tensor of shape `[batch_size]`.
            Time steps beyond the respective sequence lengths will be masked.
        discount (float): A scalar. The discount factor.

    Returns:
        A 2D Tensor of the discounted reward.
    """
    dtype: torch.dtype = reward.dtype
    if sequence_length is not None:
        reward = mask_sequences(reward, sequence_length, dtype=dtype)

    if discount == 1.:
        reward = torch.flip(reward, (1,))
        disc_reward = torch.cumsum(reward, dim=1)
        disc_reward = torch.flip(disc_reward, (1,))
    else:
        # [max_time, batch_size]
        rev_reward_T = torch.flip(reward, (1,)).permute(1, 0)

        res = []
        acc = torch.zeros_like(reward[:, 1])
        for i in range(rev_reward_T.shape[0]):
            cur = rev_reward_T[i]
            acc = cur + discount * acc
            res.append(acc)

        rev_reward_T_cum = torch.stack(res, dim=0)
        disc_reward = torch.flip(rev_reward_T_cum.permute(1, 0), (1,))

    return disc_reward



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for embedder utils.
"""

import unittest

from texar.torch.modules.embedders import embedder_utils


class GetEmbeddingTest(unittest.TestCase):
    """Tests embedding creator.
    """

    def test_get_embedding(self):
        """Tests :func:`~texar.torch.modules.embedder.embedder_utils.get_embedding`.
        """
        vocab_size = 100
        emb = embedder_utils.get_embedding(num_embeds=vocab_size)
        self.assertEqual(emb.size(0), vocab_size)
        self.assertEqual(emb.size(1),
                         embedder_utils.default_embedding_hparams()["dim"])

        hparams = {
            "initializer": {
                "type": "torch.nn.init.uniform_",
                "kwargs": {'a': -0.1, 'b': 0.1}
            }
        }
        emb = embedder_utils.get_embedding(
            hparams=hparams, num_embeds=vocab_size, )
        self.assertEqual(emb.size(0), vocab_size)
        self.assertEqual(emb.size(1),
                         embedder_utils.default_embedding_hparams()["dim"])


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Pre-trained GPT-2 tokenizer.

Code structure adapted from:
    `https://github.com/huggingface/pytorch-transformers/blob/master/pytorch_transformers/tokenization_gpt2.py`
"""

from typing import Any, Dict, List, Optional, Tuple

import os
import json
import regex as re

from texar.torch.modules.pretrained.gpt2 import PretrainedGPT2Mixin
from texar.torch.data.tokenizers.tokenizer_base import TokenizerBase
from texar.torch.data.tokenizers.gpt2_tokenizer_utils import \
    bytes_to_unicode, get_pairs

__all__ = [
    'GPT2Tokenizer',
]


class GPT2Tokenizer(TokenizerBase, PretrainedGPT2Mixin):
    r"""Pre-trained GPT2 Tokenizer.

    Args:
        pretrained_model_name (optional): a `str`, the name of
            pre-trained model (e.g., `117M`). Please refer to
            :class:`~texar.torch.modules.PretrainedGPT2Mixin` for
            all supported models.
            If None, the model name in :attr:`hparams` is used.
        cache_dir (optional): the path to a folder in which the
            pre-trained models will be cached. If `None` (default),
            a default directory (``texar_data`` folder under user's home
            directory) will be used.
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparameter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure
            and default values.
    """

    _IS_PRETRAINED = True
    _MAX_INPUT_SIZE = {
        'gpt2-small': 1024,
        'gpt2-medium': 1024,
        'gpt2-large': 1024,
        'gpt2-xl': 1024,
    }
    _DEPRECATED_MAX_INPUT_SIZE = {
        '117M': 1024,
        '345M': 1024,
    }
    _MAX_INPUT_SIZE.update(_DEPRECATED_MAX_INPUT_SIZE)

    _VOCAB_FILE_NAMES = {
        'vocab_file': 'encoder.json',
        'merges_file': 'vocab.bpe',
    }
    _VOCAB_FILE_MAP = {
        'vocab_file': {
            'gpt2-small': 'encoder.json',
            'gpt2-medium': 'encoder.json',
            'gpt2-large': 'encoder.json',
            'gpt2-xl': 'encoder.json',
            '117M': 'encoder.json',
            '345M': 'encoder.json',
        },
        'merges_file': {
            'gpt2-small': 'vocab.bpe',
            'gpt2-medium': 'vocab.bpe',
            'gpt2-large': 'vocab.bpe',
            'gpt2-xl': 'vocab.bpe',
            '117M': 'vocab.bpe',
            '345M': 'vocab.bpe',
        },
    }

    def __init__(self,
                 pretrained_model_name: Optional[str] = None,
                 cache_dir: Optional[str] = None,
                 hparams=None):
        self.load_pretrained_config(pretrained_model_name, cache_dir, hparams)

        super().__init__(hparams=None)

        self.config = {
            'errors': self.hparams['errors']
        }

        if self.pretrained_model_dir is not None:
            assert self.pretrained_model_name is not None
            vocab_file = os.path.join(self.pretrained_model_dir,
                                      self._VOCAB_FILE_MAP['vocab_file']
                                      [self.pretrained_model_name])
            merges_file = os.path.join(self.pretrained_model_dir,
                                       self._VOCAB_FILE_MAP['merges_file']
                                       [self.pretrained_model_name])
            assert pretrained_model_name is not None
            if self._MAX_INPUT_SIZE.get(pretrained_model_name):
                self.max_len = self._MAX_INPUT_SIZE[pretrained_model_name]
        else:
            vocab_file = self.hparams['vocab_file']
            merges_file = self.hparams['merges_file']
            if self.hparams.get('max_len'):
                self.max_len = self.hparams['max_len']

        if not os.path.isfile(vocab_file):
            raise ValueError("Can't find a vocabulary file at path "
                             "'{}".format(vocab_file))

        if not os.path.isfile(merges_file):
            raise ValueError("Can't find a merges file at path "
                             "'{}".format(merges_file))

        with open(vocab_file) as fp:
            self.encoder = json.load(fp)
        self.decoder = {v: k for k, v in self.encoder.items()}
        self.errors = self.hparams["errors"]  # how to handle errors in decoding
        self.byte_encoder = bytes_to_unicode()
        self.byte_decoder = {v: k for k, v in self.byte_encoder.items()}
        with open(merges_file, encoding='utf-8') as fp:
            bpe_data = fp.read().split('\n')[1:-1]
        bpe_merges = [tuple(merge.split()) for merge in bpe_data]
        self.bpe_ranks = dict(zip(bpe_merges, range(len(bpe_merges))))
        self.cache: Dict[str, str] = {}

        # Should haved added re.IGNORECASE so BPE merges can happen for
        # capitalized versions of contractions
        self.pat = re.compile(r"""'s|'t|'re|'ve|'m|'ll|'d| ?\p{L}+| """ +
                              r""""?\p{N}+| ?[^\s\p{L}\p{N}]+|\s+(?!\S)|\s+""")

    def _map_text_to_token(self, text: str) -> List[str]:  # type: ignore
        r"""Tokenize a string. """
        bpe_tokens: List[str] = []
        for token in re.findall(self.pat, text):
            token = ''.join(self.byte_encoder[b] for b in token.encode('utf-8'))
            bpe_tokens.extend(
                bpe_token for bpe_token in self._bpe(token).split(' '))
        return bpe_tokens

    def save_vocab(self, save_dir: str) -> Tuple[str, str]:
        r"""Save the tokenizer vocabulary and merge files to a directory."""
        if not os.path.isdir(save_dir):
            raise ValueError("Vocabulary path ({}) should be a "
                             "directory".format(save_dir))

        vocab_file = os.path.join(save_dir,
                                  self._VOCAB_FILE_NAMES['vocab_file'])
        merge_file = os.path.join(save_dir,
                                  self._VOCAB_FILE_NAMES['merges_file'])

        with open(vocab_file, 'w', encoding='utf-8') as f:
            f.write(json.dumps(self.encoder, ensure_ascii=False))

        index = 0
        with open(merge_file, "w", encoding="utf-8") as writer:
            writer.write(u'#version: 0.2\n')
            for bpe_tokens, token_index in sorted(self.bpe_ranks.items(),
                                                  key=lambda kv: kv[1]):
                if index != token_index:
                    print("Saving vocabulary to {}: BPE merge indices are "
                          "not consecutive. Please check that the tokenizer "
                          "is not corrupted!".format(merge_file))
                    index = token_index
                writer.write(' '.join(bpe_tokens) + u'\n')
                index += 1

        return (vocab_file, merge_file)

    def _bpe(self, token: str) -> str:
        if token in self.cache:
            return self.cache[token]
        word = tuple(token)
        pairs = get_pairs(word)

        if not pairs:
            return token

        while True:
            bigram = min(pairs, key=lambda pair: self.bpe_ranks.get(
                pair, float('inf')))
            if bigram not in self.bpe_ranks:
                break
            first, second = bigram
            new_word: List[str] = []
            i = 0
            while i < len(word):
                try:
                    j = word.index(first, i)
                    new_word.extend(word[i:j])
                    i = j
                except ValueError:
                    new_word.extend(word[i:])
                    break

                if word[i] == first and i < len(word) - 1 \
                        and word[i + 1] == second:
                    new_word.append(first + second)
                    i += 2
                else:
                    new_word.append(word[i])
                    i += 1
            new_word = tuple(new_word)
            word = new_word
            if len(word) == 1:
                break
            pairs = get_pairs(word)
        word = ' '.join(word)
        self.cache[token] = word
        return word

    @property
    def vocab_size(self) -> int:
        return len(self.encoder)

    def _map_token_to_id(self, token: str) -> int:
        r"""Maps a token to an id using the vocabulary."""
        return self.encoder.get(token, self.encoder.get(self.unk_token))

    def _map_id_to_token(self, index: int) -> str:
        r"""Maps an id to a token using the vocabulary."""
        token = self.decoder.get(index)
        assert isinstance(token, str)
        return token

    def map_token_to_text(self, tokens: List[str]) -> str:
        r"""Maps a sequence of tokens (string) in a single string."""
        text = ''.join(tokens)
        text = bytearray([self.byte_decoder[c] for c in text]).decode(
            'utf-8', errors=self.errors)
        return text

    def encode_text(  # type: ignore
            self,
            text: str,
            max_seq_length: Optional[int] = None,
            append_eos_token: bool = True) -> Tuple[List[int], int]:
        r"""Adds special tokens to a sequence and computes the corresponding
        sequence length for GPT2 specific tasks. The sequence will be truncated
        if its length is larger than ``max_seq_length``.

        A GPT2 sequence has the following format:
        `[bos_token]` X `[eos_token]` `[pad_token]`

        Args:
            text: Input text.
            max_seq_length: Maximum sequence length.
            append_eos_token: Whether to append ``eos_token`` after the
                sequence.

        Returns:
            A tuple of `(input_ids, seq_len)`, where

            - ``input_ids``: A list of input token ids with added
              special tokens.
            - ``seq_len``: The sequence length.
        """
        if max_seq_length is None:
            max_seq_length = self.max_len

        token_ids = self.map_text_to_id(text)
        assert isinstance(token_ids, list)

        bos_token_id = self._map_token_to_id(self.bos_token)
        eos_token_id = self._map_token_to_id(self.eos_token)
        pad_token_id = self._map_token_to_id(self.pad_token)

        if append_eos_token:
            input_ids = token_ids[:max_seq_length - 2]
            input_ids = [bos_token_id] + input_ids + [eos_token_id]
        else:
            input_ids = token_ids[:max_seq_length - 1]
            input_ids = [bos_token_id] + input_ids

        seq_len = len(input_ids)

        # Pad up to the maximum sequence length.
        input_ids = input_ids + [pad_token_id] * (max_seq_length - seq_len)

        assert len(input_ids) == max_seq_length

        return input_ids, seq_len

    @staticmethod
    def default_hparams() -> Dict[str, Any]:
        r"""Returns a dictionary of hyperparameters with default values.

        * The tokenizer is determined by the constructor argument
          :attr:`pretrained_model_name` if it's specified. In this case,
          `hparams` are ignored.
        * Otherwise, the tokenizer is determined by
          `hparams['pretrained_model_name']` if it's specified. All other
          configurations in `hparams` are ignored.
        * If the above two are `None`, the tokenizer is defined by the
          configurations in `hparams`.

        .. code-block:: python

            {
                "pretrained_model_name": "117M",
                "vocab_file": None,
                "merges_file": None,
                "max_len": 1024,
                "bos_token": "<|endoftext|>",
                "eos_token": "<|endoftext|>",
                "unk_token": "<|endoftext|>",
                "pad_token": "<|endoftext|>",
                "errors": "replace",
                "name": "gpt2_tokenizer",
            }

        Here:

        `"pretrained_model_name"`: str or None
            The name of the pre-trained GPT2 model.

        `"vocab_file"`: str or None
            The path to a vocabulary json file mapping tokens to ids.

        `"merges_file"`: str or None
            The path to a merges file.

        `"max_len"`: int
            The maximum sequence length that this model might ever be used with.

        `"bos_token"`: str
            Beginning of sentence token

        `"eos_token"`: str
            End of sentence token

        `"unk_token"`: str
            Unknown token

        `"pad_token"`: str
            Padding token

        `"errors"`: str
            Response when mapping tokens to text fails. The possible values are
            `ignore`, `replace`, and `strict`.

        `"name"`: str
            Name of the tokenizer.
        """
        return {
            'pretrained_model_name': '117M',
            'vocab_file': None,
            'merges_file': None,
            'max_len': 1024,
            'bos_token': '<|endoftext|>',
            'eos_token': '<|endoftext|>',
            'unk_token': '<|endoftext|>',
            'pad_token': '<|endoftext|>',
            'errors': 'replace',
            'name': 'gpt2_tokenizer',
            '@no_typecheck': ['pretrained_model_name'],
        }

    @classmethod
    def _transform_config(cls, pretrained_model_name: str,
                          cache_dir: str):
        r"""Returns the configuration of the pre-trained GPT2 tokenizer."""
        return {
            'vocab_file': None,
            'merges_file': None,
            'max_len': 1024,
            'bos_token': '<|endoftext|>',
            'eos_token': '<|endoftext|>',
            'unk_token': '<|endoftext|>',
            'pad_token': '<|endoftext|>',
            'errors': 'replace',
        }



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for SentencePiece tokenizer.
"""

import unittest

import os
import pickle
import tempfile

from texar.torch.data.data_utils import maybe_download
from texar.torch.data.tokenizers.sentencepiece_tokenizer import \
    SentencePieceTokenizer


class SentencePieceTokenizerTest(unittest.TestCase):

    def setUp(self):
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.SAMPLE_VOCAB = maybe_download(
            'https://github.com/google/sentencepiece/blob/master/'
            'python/test/test_model.model?raw=true', self.tmp_dir.name)

        self.tokenizer = SentencePieceTokenizer.load(self.SAMPLE_VOCAB)

        self.tokenizer.save(self.tmp_dir.name)

    def tearDown(self):
        self.tmp_dir.cleanup()

    def test_load(self):
        tokenizer = SentencePieceTokenizer.load(self.tmp_dir.name)

        self.assertEqual(1000, len(tokenizer))
        self.assertEqual(0, tokenizer.map_token_to_id('<unk>'))
        self.assertEqual(1, tokenizer.map_token_to_id('<s>'))
        self.assertEqual(2, tokenizer.map_token_to_id('</s>'))
        self.assertEqual('<unk>', tokenizer.map_id_to_token(0))
        self.assertEqual('<s>', tokenizer.map_id_to_token(1))
        self.assertEqual('</s>', tokenizer.map_id_to_token(2))
        for i in range(len(tokenizer)):
            token = tokenizer.map_id_to_token(i)
            self.assertEqual(i, tokenizer.map_token_to_id(token))

    def test_roundtrip(self):
        tokenizer = SentencePieceTokenizer.load(self.tmp_dir.name)

        text = 'I saw a girl with a telescope.'
        ids = tokenizer.map_text_to_id(text)
        tokens = tokenizer.map_text_to_token(text)

        self.assertEqual(text, tokenizer.map_id_to_text(ids))
        self.assertEqual(text, tokenizer.map_token_to_text(tokens))

    def test_train(self):
        tmp_dir = tempfile.TemporaryDirectory()
        TEXT_FILE = maybe_download(
            'https://github.com/google/sentencepiece/blob/master/'
            'data/botchan.txt?raw=true', tmp_dir.name)

        hparams = {
            "vocab_file": None,
            "text_file": TEXT_FILE,
            "vocab_size": 1000,
        }
        tokenizer = SentencePieceTokenizer(hparams=hparams)

        with open(TEXT_FILE, 'r', encoding='utf-8') as file:
            for line in file:
                tokenizer.map_token_to_text(tokenizer.map_text_to_token(line))
                tokenizer.map_id_to_text(tokenizer.map_text_to_id(line))

    def test_pickle(self):
        tokenizer = SentencePieceTokenizer.load(self.tmp_dir.name)
        self.assertIsNotNone(tokenizer)

        text = u"Munich and Berlin are nice cities"
        subwords = tokenizer.map_text_to_token(text)

        with tempfile.TemporaryDirectory() as tmpdirname:
            filename = os.path.join(tmpdirname, u"tokenizer.bin")
            with open(filename, "wb") as f:
                pickle.dump(tokenizer, f)
            with open(filename, "rb") as f:
                tokenizer_new = pickle.load(f)

        subwords_loaded = tokenizer_new.map_text_to_token(text)

        self.assertListEqual(subwords, subwords_loaded)

    def test_save_load(self):
        tokenizer = SentencePieceTokenizer.load(self.tmp_dir.name)

        before_tokens = tokenizer.map_text_to_id(
            u"He is very happy, UNwant\u00E9d,running")

        with tempfile.TemporaryDirectory() as tmpdirname:
            tokenizer.save(tmpdirname)
            tokenizer = tokenizer.load(tmpdirname)

        after_tokens = tokenizer.map_text_to_id(
            u"He is very happy, UNwant\u00E9d,running")
        self.assertListEqual(before_tokens, after_tokens)

    def test_add_tokens(self):
        tokenizer = SentencePieceTokenizer.load(self.tmp_dir.name)

        vocab_size = tokenizer.vocab_size
        all_size = len(tokenizer)

        self.assertNotEqual(vocab_size, 0)
        self.assertEqual(vocab_size, all_size)

        new_toks = ["aaaaabbbbbb", "cccccccccdddddddd"]
        added_toks = tokenizer.add_tokens(new_toks)
        vocab_size_2 = tokenizer.vocab_size
        all_size_2 = len(tokenizer)

        self.assertNotEqual(vocab_size_2, 0)
        self.assertEqual(vocab_size, vocab_size_2)
        self.assertEqual(added_toks, len(new_toks))
        self.assertEqual(all_size_2, all_size + len(new_toks))

        tokens = tokenizer.map_text_to_id("aaaaabbbbbb low cccccccccdddddddd l")
        self.assertGreaterEqual(len(tokens), 4)
        self.assertGreater(tokens[0], tokenizer.vocab_size - 1)
        self.assertGreater(tokens[-2], tokenizer.vocab_size - 1)

        new_toks_2 = {'eos_token': ">>>>|||<||<<|<<",
                      'pad_token': "<<<<<|||>|>>>>|>"}
        added_toks_2 = tokenizer.add_special_tokens(new_toks_2)
        vocab_size_3 = tokenizer.vocab_size
        all_size_3 = len(tokenizer)

        self.assertNotEqual(vocab_size_3, 0)
        self.assertEqual(vocab_size, vocab_size_3)
        self.assertEqual(added_toks_2, len(new_toks_2))
        self.assertEqual(all_size_3, all_size_2 + len(new_toks_2))

        tokens = tokenizer.map_text_to_id(
            ">>>>|||<||<<|<< aaaaabbbbbb low cccccccccdddddddd "
            "<<<<<|||>|>>>>|> l")

        self.assertGreaterEqual(len(tokens), 6)
        self.assertGreater(tokens[0], tokenizer.vocab_size - 1)
        self.assertGreater(tokens[0], tokens[1])
        self.assertGreater(tokens[-2], tokenizer.vocab_size - 1)
        self.assertGreater(tokens[-2], tokens[-3])
        self.assertEqual(tokens[0],
                         tokenizer.map_token_to_id(tokenizer.eos_token))
        self.assertEqual(tokens[-2],
                         tokenizer.map_token_to_id(tokenizer.pad_token))


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various neural networks and related utilities.
"""

from texar.torch.modules.networks.network_base import FeedForwardNetworkBase
from texar.torch.utils.utils import get_output_size

__all__ = [
    "FeedForwardNetwork",
]


class FeedForwardNetwork(FeedForwardNetworkBase):
    r"""Feed-forward neural network that consists of a sequence of layers.

    Args:
        layers (list, optional): A list of :torch_nn:`Linear`
            instances composing the network. If not given, layers are created
            according to :attr:`hparams`.
        hparams (dict, optional): Embedder hyperparameters. Missing
            hyperparameters will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.

    See :meth:`forward` for the inputs and outputs.

    Example:

        .. code-block:: python

            hparams = { # Builds a two-layer dense NN
                "layers": [
                    { "type": "Dense", "kwargs": { "units": 256 },
                    { "type": "Dense", "kwargs": { "units": 10 }
                ]
            }
            nn = FeedForwardNetwork(hparams=hparams)

            inputs = torch.randn([64, 100])
            outputs = nn(inputs)
            # outputs == Tensor of shape [64, 10]
    """

    def __init__(self, layers=None, hparams=None):
        super().__init__(hparams=hparams)

        self._build_layers(layers=layers, layer_hparams=self._hparams.layers)

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "layers": [],
                "name": "NN"
            }

        Here:

        `"layers"`: list
            A list of layer hyperparameters. See
            :func:`~texar.torch.core.get_layer` for details on layer
            hyperparameters.

        `"name"`: str
            Name of the network.
        """
        return {
            "layers": [],
            "name": "NN"
        }

    @property
    def output_size(self) -> int:
        r"""The feature size of network layers output. If output size is
        only determined by input, the feature size is equal to ``-1``.
        """
        for i, layer in enumerate(reversed(self._layers)):
            size = get_output_size(layer)
            size_ext = getattr(layer, 'output_size', None)
            if size_ext is not None:
                size = size_ext
            if size is None:
                break
            if size > 0:
                return size
            elif i == len(self._layers) - 1:
                return -1

        raise ValueError("'output_size' can not be calculated because "
                         "'FeedForwardNetwork' contains submodule "
                         "whose output size cannot be determined.")



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Attentions specific to Transformer.
"""

from typing import Optional, Tuple

import numpy as np
import torch

__all__ = [
    'attention_bias_lower_triangle',
    'attention_bias_ignore_padding',
    'attention_bias_local',
]


def attention_bias_lower_triangle(length: int,
                                  bias_value: float = -1e18) -> torch.Tensor:
    r"""Create an bias tensor to be added to attention logits.
    Allows a query to attend to all positions up to and including its own.

    Args:
        length: a scalar.
        bias_value: value to fill the bias tensor with.

    Returns:
        a ``Tensor`` with shape [1, 1, length, length].
    """
    return attention_bias_local(length, -1, 0, bias_value)


def attention_bias_local(length: int, max_backward: int, max_forward: int,
                         bias_value: float = -1e18) -> torch.Tensor:
    r"""Create an bias tensor to be added to attention logits.
    A position may attend to positions at most max_distance from it,
    forward and backwards.

    This does not actually save any computation.

    Args:
        length: int
        max_backward: int, maximum distance backward to attend. Negative
            values indicate unlimited.
        max_forward: int, maximum distance forward to attend. Negative
            values indicate unlimited.
        bias_value: value to fill the bias tensor with.

    Returns:
        a ``Tensor`` with shape [1, 1, length, length].
        [batch_size, num_heads, query_len, query_len]
    """
    band = _ones_matrix_band_part(
        length,
        length,
        max_backward,
        max_forward,
        out_shape=(1, 1, length, length))
    return bias_value * (1.0 - band)


def attention_bias_ignore_padding(memory_padding: torch.Tensor,
                                  bias_value: float = -1e18) -> torch.Tensor:
    r"""Create an bias tensor to be added to attention logits.

    Args:
        memory_padding: a float ``Tensor`` with shape [batch, memory_length].
        bias_value: value to fill the bias tensor with.

    Returns:
        a ``Tensor`` with shape [batch, 1, 1, memory_length].
        each dim corresponding to batch_size, num_heads, queries_len,
        memory_length
    """
    ret = memory_padding * bias_value
    return ret.view(ret.size(0), 1, 1, ret.size(-1))


def _ones_matrix_band_part(rows: int, cols: int, num_lower: int, num_upper: int,
                           out_shape: Optional[Tuple[int, ...]] = None) \
        -> torch.Tensor:
    r"""Matrix band part of ones.
    """
    if num_lower < 0:
        num_lower = rows - 1
    if num_upper < 0:
        num_upper = cols - 1
    lower_mask = np.tri(cols, rows, num_lower).T
    upper_mask = np.tri(rows, cols, num_upper)
    band = np.ones((rows, cols)) * lower_mask * upper_mask
    if out_shape:
        band = band.reshape(out_shape)
    band = torch.as_tensor(band, dtype=torch.float32)
    return band



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Pre-trained T5 tokenizer.
"""

from typing import Any, Dict, Optional

import os
import re

from texar.torch.data.tokenizers.sentencepiece_tokenizer \
    import SentencePieceTokenizer
from texar.torch.modules.pretrained.t5 import PretrainedT5Mixin

__all__ = [
    'T5Tokenizer',
]


class T5Tokenizer(SentencePieceTokenizer, PretrainedT5Mixin):
    r"""Pre-trained T5 Tokenizer.

    Args:
        pretrained_model_name (optional): a `str`, the name of
            pre-trained model (e.g., `T5-Small`). Please refer to
            :class:`~texar.torch.modules.PretrainedT5Mixin` for
            all supported models.
            If None, the model name in :attr:`hparams` is used.
        cache_dir (optional): the path to a folder in which the
            pre-trained models will be cached. If `None` (default),
            a default directory (``texar_data`` folder under user's home
            directory) will be used.
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparameters will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure
            and default values.
    """

    _IS_PRETRAINED = True

    _VOCAB_FILE_NAMES = {
        'vocab_file': 'sentencepiece.model'
    }

    _MAX_INPUT_SIZE = {
        'T5-Small': 512,
        'T5-Base': 512,
        'T5-Large': 512,
        'T5-3B': 512,
        'T5-11B': 512
    }

    def __init__(self,
                 pretrained_model_name: Optional[str] = None,
                 cache_dir: Optional[str] = None,
                 hparams=None):

        self.load_pretrained_config(pretrained_model_name, cache_dir, hparams)

        if self.pretrained_model_dir is not None:
            assert self.pretrained_model_name is not None
            vocab_file = os.path.join(self.pretrained_model_dir,
                                      self._VOCAB_FILE_NAMES['vocab_file'])

            if self._MAX_INPUT_SIZE.get(self.pretrained_model_name):
                self.max_len = self._MAX_INPUT_SIZE[self.pretrained_model_name]
            setattr(self.hparams, 'vocab_file', vocab_file)
        else:
            if self.hparams.get('max_len'):
                self.max_len = self.hparams['max_len']

        # Add extra_ids to the special token list
        additional_special_tokens = []
        extra_ids = self.hparams['extra_ids']
        if extra_ids > 0:
            additional_special_tokens.extend(
                ["<extra_id_{}>".format(i) for i in range(extra_ids)])

        setattr(self.hparams, 'additional_special_tokens',
                additional_special_tokens)

        super().__init__(hparams=None)

    @staticmethod
    def default_hparams() -> Dict[str, Any]:
        r"""Returns a dictionary of hyperparameters with default values.

        * The tokenizer is determined by the constructor argument
          :attr:`pretrained_model_name` if it's specified. In this case,
          `hparams` are ignored.
        * Otherwise, the tokenizer is determined by
          `hparams['pretrained_model_name']` if it's specified. All other
          configurations in `hparams` are ignored.
        * If the above two are `None`, the tokenizer is defined by the
          configurations in `hparams`.

        .. code-block:: python

            {
                "pretrained_model_name": "T5-Small",
                "vocab_file": None,
                "max_len": 512,
                "bos_token": None,
                "eos_token": "</s>",
                "unk_token": "<unk>",
                "pad_token": "<pad>",
                "extra_ids": 100,
                "additional_special_tokens": [],
                "name": "t5_tokenizer",
            }

        Here:

        `"pretrained_model_name"`: str or None
            The name of the pre-trained T5 model.

        `"vocab_file"`: str or None
            The path to a sentencepiece vocabulary file.

        `"max_len"`: int or None
            The maximum sequence length that this model might ever be used with.

        `"bos_token"`: str or None
            Beginning of sentence token. Set None to disable ``bos_token``.

        `"eos_token"`: str
            End of sentence token. Set None to disable ``eos_token``.

        `"unk_token"`: str
            Unknown token. Set None to disable ``unk_token``.

        `"pad_token"`: str
            Padding token. Set None to disable ``pad_token``.

        `"extra_ids"`: int
            Add a number of extra ids added to the end of the vocabulary for
            use as sentinels. These tokens are accessible as `<extra_id_{%d}>`
            where `{%d}` is a number between 0 and extra_ids-1. Extra tokens
            are indexed from the end of the vocabulary up to beginning
            (<extra_id_0> is the last token in the vocabulary) (like in T5
            preprocessing) see:
            `https://github.com/google-research/text-to-text-transfer-transformer/blob/9fd7b14a769417be33bc6c850f9598764913c833/t5/data/preprocessors.py#L2117`

        `"additional_special_tokens"`: list
            A list of additional special tokens.

        `"name"`: str
            Name of the tokenizer.
        """
        return {
            'pretrained_model_name': 'T5-Small',
            'vocab_file': None,
            'max_len': 512,
            'bos_token': None,
            'eos_token': '</s>',
            'unk_token': '<unk>',
            'pad_token': '<pad>',
            'extra_ids': 100,
            'additional_special_tokens': [],
            'name': 't5_tokenizer',
            '@no_typecheck': ['pretrained_model_name'],
        }

    @property
    def vocab_size(self) -> int:
        return len(self.sp_model) + self.hparams['extra_ids']

    def _map_token_to_id(self, token: str) -> int:
        if token.startswith("<extra_id_"):
            match = re.match(r"<extra_id_(\d+)>", token)
            num = int(match.group(1))  # type: ignore
            return self.vocab_size - num - 1
        return self.sp_model.PieceToId(token)

    def _map_id_to_token(self, index: int) -> str:
        if index < self.sp_model.get_piece_size():
            token = self.sp_model.IdToPiece(index)
        else:
            token = "<extra_id_{}>".format(self.vocab_size - 1 - index)
        return token



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for RNN encoders.
"""

import unittest

import torch

from texar.torch.modules.encoders.rnn_encoders import (
    UnidirectionalRNNEncoder, BidirectionalRNNEncoder)


class UnidirectionalRNNEncoderTest(unittest.TestCase):
    r"""Tests unidirectional rnn encoder.
    """

    def setUp(self):
        self._batch_size = 16
        self._max_time = 8
        self._input_size = 10

    def test_trainable_variables(self):
        r"""Tests the functionality of automatically collecting trainable
        variables.
        """
        inputs = torch.rand(self._batch_size, self._max_time, self._input_size)

        # case 1
        encoder = UnidirectionalRNNEncoder(input_size=self._input_size)
        output, _ = encoder(inputs)
        self.assertEqual(len(encoder.trainable_variables), 4)
        self.assertEqual(output.size()[-1], encoder.output_size)

        # case 2
        hparams = {
            "rnn_cell": {
                "dropout": {
                    "input_keep_prob": 0.5
                }
            }
        }
        encoder = UnidirectionalRNNEncoder(input_size=self._input_size,
                                           hparams=hparams)
        output, _ = encoder(inputs)
        self.assertEqual(len(encoder.trainable_variables), 4)
        self.assertEqual(output.size()[-1], encoder.output_size)

        # case 3
        hparams = {"output_layer": {
            "num_layers": 2,
            "layer_size": [100, 6],
            "activation": "ReLU",
            "final_layer_activation": "Identity",
            "dropout_layer_ids": [0, 1, 2],
            "variational_dropout": False}}

        encoder = UnidirectionalRNNEncoder(input_size=self._input_size,
                                           hparams=hparams)

        output, _ = encoder(inputs)
        self.assertEqual(len(encoder.trainable_variables), 8)
        self.assertEqual(output.size()[-1], encoder.output_size)

    def test_encode(self):
        r"""Tests encoding.
        """
        inputs = torch.rand(self._batch_size, self._max_time, self._input_size)

        # case 1
        encoder = UnidirectionalRNNEncoder(input_size=self._input_size)
        outputs, state = encoder(inputs)

        cell_dim = encoder.hparams.rnn_cell.kwargs.num_units

        self.assertEqual(outputs.shape, torch.Size([self._batch_size,
                                                    self._max_time,
                                                    cell_dim]))
        self.assertIsInstance(state, tuple)
        self.assertEqual(state[0].shape, torch.Size([self._batch_size,
                                                     cell_dim]))
        self.assertEqual(state[1].shape, torch.Size([self._batch_size,
                                                     cell_dim]))

        # case 2: with output layers
        hparams = {
            "output_layer": {
                "num_layers": 2,
                "layer_size": [100, 6],
                "dropout_layer_ids": [0, 1, 2],
                "variational_dropout": False
            }
        }
        encoder = UnidirectionalRNNEncoder(input_size=self._input_size,
                                           hparams=hparams)
        outputs, state, cell_outputs, output_size = encoder(
            inputs, return_cell_output=True, return_output_size=True)

        self.assertEqual(output_size, 6)
        self.assertEqual(cell_outputs.shape[-1], encoder.cell.hidden_size)

        out_dim = encoder.hparams.output_layer.layer_size[-1]
        self.assertEqual(outputs.shape, torch.Size([self._batch_size,
                                                    self._max_time,
                                                    out_dim]))


class BidirectionalRNNEncoderTest(unittest.TestCase):
    r"""Tests bidirectional rnn encoder.
    """

    def setUp(self):
        self._batch_size = 16
        self._max_time = 8
        self._input_size = 10

    def test_trainable_variables(self):
        r"""Tests the functionality of automatically collecting trainable
        variables.
        """
        inputs = torch.rand(self._batch_size, self._max_time, self._input_size)

        # case 1
        encoder = BidirectionalRNNEncoder(input_size=self._input_size)
        outputs, _ = encoder(inputs)
        self.assertEqual(len(encoder.trainable_variables), 8)
        self.assertEqual(
            (outputs[0].size(-1), outputs[1].size(-1)), (encoder.output_size))

        # case 2
        hparams = {
            "rnn_cell_fw": {
                "dropout": {
                    "input_keep_prob": 0.5
                }
            }
        }
        encoder = BidirectionalRNNEncoder(input_size=self._input_size,
                                          hparams=hparams)
        outputs, _ = encoder(inputs)
        self.assertEqual(len(encoder.trainable_variables), 8)
        self.assertEqual(
            (outputs[0].size(-1), outputs[1].size(-1)), (encoder.output_size))

        # case 3
        hparams = {
            "output_layer_fw": {
                "num_layers": 2,
                "layer_size": [100, 6],
                "activation": "ReLU",
                "final_layer_activation": "Identity",
                "dropout_layer_ids": [0, 1, 2],
                "variational_dropout": False
            },
            "output_layer_bw": {
                "num_layers": 3,
                "other_dense_kwargs": {"bias": False}
            },
            "output_layer_share_config": False
        }
        encoder = BidirectionalRNNEncoder(input_size=self._input_size,
                                          hparams=hparams)
        outputs, _ = encoder(inputs)
        self.assertEqual(len(encoder.trainable_variables), 8 + 3 + 4)
        self.assertEqual(
            (outputs[0].size(-1), outputs[1].size(-1)), (encoder.output_size))

    def test_encode(self):
        r"""Tests encoding.
        """
        inputs = torch.rand(self._batch_size, self._max_time, self._input_size)

        # case 1
        encoder = BidirectionalRNNEncoder(input_size=self._input_size)
        outputs, state = encoder(inputs)

        cell_dim = encoder.hparams.rnn_cell_fw.kwargs.num_units

        self.assertIsInstance(outputs, tuple)
        self.assertEqual(outputs[0].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       cell_dim]))
        self.assertEqual(outputs[1].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       cell_dim]))

        self.assertIsInstance(state, tuple)
        self.assertIsInstance(state[0], tuple)
        self.assertEqual(state[0][0].shape, torch.Size([self._batch_size,
                                                        cell_dim]))
        self.assertEqual(state[0][0].shape, torch.Size([self._batch_size,
                                                        cell_dim]))
        self.assertIsInstance(state[1], tuple)
        self.assertEqual(state[1][0].shape, torch.Size([self._batch_size,
                                                        cell_dim]))
        self.assertEqual(state[1][0].shape, torch.Size([self._batch_size,
                                                        cell_dim]))

        # case 2:
        hparams = {
            "output_layer_fw": {
                "num_layers": 2,
                "layer_size": [100, 6],
                "dropout_layer_ids": [0, 1, 2],
                "variational_dropout": False
            }
        }
        encoder = BidirectionalRNNEncoder(input_size=self._input_size,
                                          hparams=hparams)
        outputs, state, cell_outputs, output_size = encoder(
            inputs, return_cell_output=True, return_output_size=True)

        self.assertEqual(output_size[0], 6)
        self.assertEqual(output_size[1], 6)
        self.assertEqual(cell_outputs[0].shape[-1], encoder.cell_fw.hidden_size)
        self.assertEqual(cell_outputs[1].shape[-1], encoder.cell_bw.hidden_size)

        out_dim = encoder.hparams.output_layer_fw.layer_size[-1]

        self.assertEqual(outputs[0].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       out_dim]))
        self.assertEqual(outputs[1].shape, torch.Size([self._batch_size,
                                                       self._max_time,
                                                       out_dim]))


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Base classes for Executor metrics.
"""

from abc import ABC, abstractmethod
from typing import Generic, List, Optional, Sequence, TypeVar

__all__ = [
    "Metric",
    "SimpleMetric",
    "StreamingMetric",
]

Input = TypeVar('Input')
Value = TypeVar('Value')


class Metric(Generic[Input, Value], ABC):
    r"""Base class of all metrics. You should not directly inherit this class,
    but inherit from :class:`SimpleMetric` or :class:`StreamingMetric` instead.

    Subclasses can override the class attributes to indicate their behaviors:

    - :attr:`higher_is_better`: If `True`, higher (comparison using the greater
      than operator ``>`` returns `True`) values are considered better metric
      values. If `False`, lower values are considered better. Defaults to
      `True`.
    - :attr:`required_pred`: If `True`, predicted values are required to compute
      the metric value. Defaults to `True`.
    - :attr:`requires_label`: If `True`, labels are required to compute the
      metric value. Defaults to `True`.

    Keyword Args:
        pred_name (str, optional): Name of the predicted value. This will be
            used as the key to the dictionary returned by the model.
        label_name (str, optional): Name of the label. This will be used as the
            key to the batch object returned by the dataset. Defaults to
            ``"label"``.
        higher_is_better (bool, optional): If specified, the
            :attr:`higher_is_better` attribute for the instance is overwritten
            by the specified value.
    """
    higher_is_better: bool = True
    requires_pred: bool = True
    requires_label: bool = True

    def __init__(self, *, pred_name: Optional[str],
                 label_name: Optional[str] = "label",
                 higher_is_better: Optional[bool] = None):
        self.reset()
        if self.requires_pred and pred_name is None:
            raise ValueError(f"Metric {self.metric_name} requires a "
                             f"prediction name, but None is provided")
        if self.requires_label and label_name is None:
            raise ValueError(f"Metric {self.metric_name} requires a "
                             f"label name, but None is provided")
        if higher_is_better is not None:
            self.higher_is_better = higher_is_better
        if not self.requires_pred:
            pred_name = None
        if not self.requires_label:
            label_name = None
        self._pred_name = pred_name
        self._label_name = label_name

    @property
    def metric_name(self) -> str:
        r"""Name of the metric. By default, the class name is used."""
        return self.__class__.__name__

    @property
    def pred_name(self) -> Optional[str]:
        r"""Name of the predicted value. This will be used as the key to the
        dictionary returned by the model.
        """
        return self._pred_name

    @property
    def label_name(self) -> Optional[str]:
        r"""Name of the label (ground truth / gold value). This will be used as
        the key to the batch object returned by the dataset.
        """
        return self._label_name

    @abstractmethod
    def reset(self) -> None:
        r"""Reset the internal state of the metric, and erase all previously
        added data points.
        """
        raise NotImplementedError

    @abstractmethod
    def add(self, predicted: Sequence[Input], labels: Sequence[Input]) -> None:
        r"""Record a data batch in the metric.

        Args:
            predicted: The list of predicted values.
            labels: The list of labels.
        """
        raise NotImplementedError

    @abstractmethod
    def value(self) -> Value:
        r"""Compute the metric value.

        Returns:
            The metric value.
        """
        raise NotImplementedError

    def better(self, cur: Value, prev: Value) -> Optional[bool]:
        r"""Compare two metric values and return which is better.

        Args:
            cur: The "current" metric value.
            prev: The "previous" metric value.

        Returns:
            Return value is either a `bool` or `None`.

            - If `True`, the current metric value is considered better.
            - If `False`, the previous metric value is considered better.
            - If `None`, the two values are considered to be the same, or
              uncomparable.
        """
        result = (True if cur > prev else  # type: ignore
                  False if cur < prev else None)  # type: ignore
        if not self.higher_is_better and result is not None:
            result = not result
        return result


class SimpleMetric(Metric[Input, Value], ABC):
    r"""Base class of simple metrics. Simple metrics are metrics that do not
    support incremental computation. The value of the metric is computed only
    after all data points have been added.

    The default implementation of :meth:`add` simply stores the predicted values
    and labels into lists.
    """
    labels: List[Input]
    predicted: List[Input]
    _cached_value: Optional[Value]

    def reset(self) -> None:
        self.labels = []
        self.predicted = []
        self._cached_value = None

    def add(self, predicted: Sequence[Input], labels: Sequence[Input]):
        if len(predicted) != len(labels):
            raise ValueError(
                "Lists `predicted` and `labels` should have the same length")
        self.predicted.extend(predicted)
        self.labels.extend(labels)
        self._cached_value = None

    def value(self):
        if self._cached_value is not None:
            return self._cached_value
        self._cached_value = self._value()
        return self._cached_value

    def _value(self) -> Value:
        r"""Compute the metric value. This function is called in
        :meth:`texar.torch.run.metric.SimpleMetric.value` and the output is
        cached. This prevents recalculation of metrics which may be time
        consuming.

        Returns:
            The metric value.
        """
        raise NotImplementedError


class StreamingMetric(Metric[Input, Value], ABC):
    r"""Base class of streaming metrics. Streaming metrics are metrics that
    support incremental computation. The value of the metric may be queried
    before all data points have been added, and the computation should not be
    expensive.

    The default implementation of :meth:`add` only keeps track of the number of
    data points added. You should override this method.
    """
    count: int

    def reset(self) -> None:
        self.count = 0

    def add(self, predicted: Sequence[Input], labels: Sequence[Input]) -> None:
        if len(predicted) != len(labels):
            raise ValueError(
                "Lists `predicted` and `labels` should have the same length")
        self.count += len(predicted)



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit test for large file related operations.
"""
import contextlib
import resource
import time
import unittest
from typing import List, Optional, Tuple

import gc
import numpy as np
import torch

from texar.torch.data.data.data_base import DatasetBase, DataSource
from texar.torch.data.data.data_iterators import DataIterator
from texar.torch.data.data.dataset_utils import Batch
from texar.torch.data.data.text_data_base import TextLineDataSource
from texar.torch.data.vocabulary import Vocab
from texar.torch.utils.test import data_test
from texar.torch.utils.utils import AnyDict


@contextlib.contextmanager
def work_in_progress(msg):
    print(msg + "... ", flush=True)
    begin_time = time.time()
    yield
    time_consumed = time.time() - begin_time
    print(f"done. ({time_consumed:.2f}s)", flush=True)


RawExample = List[str]
Example = Tuple[np.ndarray, np.ndarray]


class ParallelData(DatasetBase[RawExample, Example]):
    def __init__(self, source: DataSource[RawExample],
                 src_vocab_path: str,
                 tgt_vocab_path: str,
                 hparams: AnyDict,
                 device: Optional[torch.device] = None):
        # hparams.update(parallelize_processing=False)
        self.src_vocab = Vocab(src_vocab_path)
        self.tgt_vocab = Vocab(tgt_vocab_path)
        self.device = device
        super().__init__(source, hparams=hparams)

    def process(self, raw_example: RawExample) -> Example:
        src, tgt = raw_example
        src = self.src_vocab.map_tokens_to_ids_py(src.split())
        tgt = self.tgt_vocab.map_tokens_to_ids_py(tgt.split())
        return src, tgt

    def collate(self, examples: List[Example]) -> Batch:
        src_pad_length = max(len(src) for src, _ in examples)
        tgt_pad_length = max(len(tgt) for _, tgt in examples)
        batch_size = len(examples)
        src_indices = np.zeros((batch_size, src_pad_length), dtype=np.int64)
        tgt_indices = np.zeros((batch_size, tgt_pad_length), dtype=np.int64)
        for b_idx, (src, tgt) in enumerate(examples):
            src_indices[b_idx, :len(src)] = src
            tgt_indices[b_idx, :len(tgt)] = tgt
        src_indices = torch.from_numpy(src_indices)
        tgt_indices = torch.from_numpy(tgt_indices)
        return Batch(batch_size, src=src_indices, tgt=tgt_indices)


def wrap_progress(func):
    from tqdm import tqdm
    return lambda: tqdm(func(), leave=False)


def get_process_memory():
    return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024 / 1024


@data_test
class LargeFileTest(unittest.TestCase):
    def setUp(self) -> None:
        self.source = TextLineDataSource(
            '../../Downloads/en-es.bicleaner07.txt.gz',
            compression_type='gzip', delimiter='\t')
        self.source.__iter__ = wrap_progress(  # type: ignore
            self.source.__iter__)
        self.num_workers = 3
        self.batch_size = 64

    def _test_modes_with_workers(self, lazy_mode: str, cache_mode: str,
                                 num_workers: int):
        from tqdm import tqdm
        gc.collect()
        mem = get_process_memory()
        with work_in_progress(f"Data loading with lazy mode '{lazy_mode}' "
                              f"and cache mode '{cache_mode}' "
                              f"with {num_workers} workers"):
            print(f"Memory before: {mem:.2f} MB")
            with work_in_progress("Construction"):
                data = ParallelData(self.source,
                                    '../../Downloads/src.vocab',
                                    '../../Downloads/tgt.vocab',
                                    {'batch_size': self.batch_size,
                                     'lazy_strategy': lazy_mode,
                                     'cache_strategy': cache_mode,
                                     'num_parallel_calls': num_workers,
                                     'shuffle': False,
                                     'allow_smaller_final_batch': False,
                                     'max_dataset_size': 100000})
            print(f"Memory after construction: {mem:.2f} MB")
            iterator = DataIterator(data)
            with work_in_progress("Iteration"):
                for batch in tqdm(iterator, leave=False):
                    self.assertEqual(batch.batch_size, self.batch_size)
            gc.collect()
            print(f"Memory after iteration: {mem:.2f} MB")
            with work_in_progress("2nd iteration"):
                for batch in tqdm(iterator, leave=False):
                    self.assertEqual(batch.batch_size, self.batch_size)

    def _test_modes(self, lazy_mode: str, cache_mode: str):
        self._test_modes_with_workers(lazy_mode, cache_mode, self.num_workers)
        self._test_modes_with_workers(lazy_mode, cache_mode, 1)

    def test_none_processed(self):
        self._test_modes('none', 'processed')

    def test_process_loaded(self):
        self._test_modes('process', 'loaded')

    def test_process_processed(self):
        self._test_modes('process', 'processed')

    def test_all_none(self):
        self._test_modes('all', 'none')

    def test_all_loaded(self):
        self._test_modes('all', 'loaded')

    def test_all_processed(self):
        self._test_modes('all', 'processed')

    def _test_all_combinations(self):
        self.test_none_processed()
        self.test_process_loaded()
        self.test_process_processed()
        self.test_all_none()
        self.test_all_loaded()
        self.test_all_processed()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for BERT classifiers.
"""

import unittest

import torch

from texar.torch.modules.classifiers.bert_classifier import *
from texar.torch.utils.test import pretrained_test


class BERTClassifierTest(unittest.TestCase):
    r"""Tests :class:`~texar.torch.modules.BERTClassifier` class.
    """

    def setUp(self) -> None:
        self.batch_size = 2
        self.max_length = 3
        self.inputs = torch.zeros(
            self.batch_size, self.max_length, dtype=torch.long)

    @pretrained_test
    def test_model_loading(self):
        r"""Tests model loading functionality."""
        for pretrained_model_name in BERTClassifier.available_checkpoints():
            classifier = BERTClassifier(
                pretrained_model_name=pretrained_model_name)
            _, _ = classifier(self.inputs)

    def test_trainable_variables(self):
        r"""Tests the functionality of automatically collecting trainable
        variables.
        """
        # case 1
        hparams = {
            "pretrained_model_name": None,
        }
        classifier = BERTClassifier(hparams=hparams)
        self.assertEqual(len(classifier.trainable_variables), 199 + 2)
        _, _ = classifier(self.inputs)

        # case 2
        hparams = {
            "pretrained_model_name": None,
            "clas_strategy": "all_time",
            "max_seq_length": 8,
        }
        classifier = BERTClassifier(hparams=hparams)
        self.assertEqual(len(classifier.trainable_variables), 199 + 2)
        _, _ = classifier(self.inputs)

        # case 3
        hparams = {
            "pretrained_model_name": None,
            "clas_strategy": "time_wise",
        }
        classifier = BERTClassifier(hparams=hparams)
        self.assertEqual(len(classifier.trainable_variables), 199 + 2)
        _, _ = classifier(self.inputs)

    def test_classification(self):
        r"""Tests classification.
        """
        inputs = torch.randint(30521, (self.batch_size, self.max_length))

        # case 1
        hparams = {
            "pretrained_model_name": None,
        }
        classifier = BERTClassifier(hparams=hparams)
        logits, preds = classifier(inputs)

        self.assertEqual(logits.shape, torch.Size(
            [self.batch_size, classifier.output_size]))
        self.assertEqual(preds.shape, torch.Size([self.batch_size]))

        # case 2
        hparams = {
            "pretrained_model_name": None,
            "num_classes": 10,
            "clas_strategy": "time_wise",
        }
        classifier = BERTClassifier(hparams=hparams)
        logits, preds = classifier(inputs)

        self.assertEqual(logits.shape, torch.Size(
            [self.batch_size, self.max_length, classifier.output_size]))
        self.assertEqual(preds.shape, torch.Size(
            [self.batch_size, self.max_length]))

        # case 3
        hparams = {
            "pretrained_model_name": None,
            "num_classes": 0,
            "clas_strategy": "time_wise",
        }
        classifier = BERTClassifier(hparams=hparams)
        logits, preds = classifier(inputs)

        self.assertEqual(logits.shape, torch.Size(
            [self.batch_size, self.max_length, classifier.output_size]))
        self.assertEqual(preds.shape, torch.Size(
            [self.batch_size, self.max_length]))

        # case 4
        hparams = {
            "pretrained_model_name": None,
            "num_classes": 10,
            "clas_strategy": "all_time",
            "max_seq_length": self.max_length,
        }
        classifier = BERTClassifier(hparams=hparams)
        logits, preds = classifier(inputs)

        self.assertEqual(logits.shape, torch.Size(
            [self.batch_size, classifier.output_size]))
        self.assertEqual(preds.shape, torch.Size([self.batch_size]))

    def test_binary(self):
        r"""Tests binary classification.
        """
        inputs = torch.randint(30521, (self.batch_size, self.max_length))

        # case 1
        hparams = {
            "pretrained_model_name": None,
            "num_classes": 1,
            "clas_strategy": "time_wise",
        }
        classifier = BERTClassifier(hparams=hparams)
        logits, preds = classifier(inputs)

        self.assertEqual(logits.shape, torch.Size(
            [self.batch_size, self.max_length]))
        self.assertEqual(preds.shape, torch.Size(
            [self.batch_size, self.max_length]))

        # case 2
        hparams = {
            "pretrained_model_name": None,
            "num_classes": 1,
            "clas_strategy": "cls_time",
            "max_seq_length": self.max_length,
        }
        classifier = BERTClassifier(hparams=hparams)
        logits, preds = classifier(inputs)

        self.assertEqual(logits.shape, torch.Size([self.batch_size]))
        self.assertEqual(preds.shape, torch.Size([self.batch_size]))

        # case 3
        hparams = {
            "pretrained_model_name": None,
            "num_classes": 1,
            "clas_strategy": "all_time",
            "max_seq_length": self.max_length,
        }
        classifier = BERTClassifier(hparams=hparams)
        logits, preds = classifier(inputs)

        self.assertEqual(logits.shape, torch.Size([self.batch_size]))
        self.assertEqual(preds.shape, torch.Size([self.batch_size]))

    def test_soft_ids(self):
        r"""Tests soft ids.
        """
        inputs = torch.rand(self.batch_size, self.max_length, 30522)

        hparams = {
            "pretrained_model_name": None,
            "num_classes": 1,
            "clas_strategy": "time_wise",
        }
        classifier = BERTClassifier(hparams=hparams)
        logits, preds = classifier(inputs)

        self.assertEqual(logits.shape, torch.Size(
            [self.batch_size, self.max_length]))
        self.assertEqual(preds.shape, torch.Size(
            [self.batch_size, self.max_length]))


if __name__ == "__main__":
    unittest.main()
